import numpy as np
from helpers.maths import Vector
from helpers.robot.control.control import Trajectory


def square(
    t_range: tuple[float, float],
    *,
    side_length: float = 1.0,
    offset: Vector = np.array([0.0, 0.5]),
) -> Trajectory:
    t_0, t_f = t_range
    x_0, y_0 = offset

    def trajectory(t: float) -> Vector:
        normalized_t = (t - t_0) / (t_f - t_0) * 4.0
        segment = int(normalized_t)
        remainder = normalized_t - segment

        match segment:
            case 0:
                return np.array([x_0 + remainder * side_length, y_0])
            case 1:
                return np.array([x_0 + side_length, y_0 + remainder * side_length])
            case 2:
                return np.array(
                    [x_0 + side_length - remainder * side_length, y_0 + side_length]
                )
            case 3:
                return np.array([x_0, y_0 + side_length - remainder * side_length])
            case _:
                return np.array([x_0, y_0])

    return trajectory


def circle(
    t_range: tuple[float, float],
    *,
    radius: float = 0.5,
    offset: Vector = np.array([0.0, 0.5]),
) -> Trajectory:
    t_0, t_f = t_range

    def trajectory(t: float) -> Vector:
        normalized_t = (t - t_0) / (t_f - t_0)
        angle = normalized_t * 2 * np.pi

        return radius * np.array([np.cos(angle), np.sin(angle)]) + offset + radius

    return trajectory


def figure_eight(
    t_range: tuple[float, float],
    *,
    radius: float = 0.5,
    offset: Vector = np.array([0.0, 0.5]),
) -> Trajectory:
    t_0, t_f = t_range

    def trajectory(t: float) -> Vector:
        normalized_t = (t - t_0) / (t_f - t_0) * 2.0 * np.pi
        x = radius * np.sin(normalized_t)
        y = radius * np.sin(2 * normalized_t)

        return np.array([x, y]) + offset + radius

    return trajectory


def spiral(
    t_range: tuple[float, float],
    *,
    radius: float = 0.5,
    offset: Vector = np.array([0.0, 0.5]),
) -> Trajectory:
    t_0, t_f = t_range

    def trajectory(t: float) -> Vector:
        normalized_t = (t - t_0) / (t_f - t_0) * 2.0 * np.pi
        x = radius * np.sin(normalized_t) * normalized_t
        y = radius * np.cos(normalized_t) * normalized_t

        return np.array([x, y]) + offset + radius

    return trajectory


def heart(
    t_range: tuple[float, float],
    *,
    radius: float = 0.5,
    offset: Vector = np.array([0.0, 0.5]),
) -> Trajectory:
    t_0, t_f = t_range

    def trajectory(t: float) -> Vector:
        normalized_t = (t - t_0) / (t_f - t_0) * 2.0 * np.pi
        x = radius * np.sin(normalized_t) * np.sqrt(np.abs(np.cos(normalized_t)))
        y = radius * np.cos(normalized_t) * np.sqrt(np.abs(np.cos(normalized_t)))

        return np.array([x, y]) + offset + radius

    return trajectory


def infinity(
    t_range: tuple[float, float],
    *,
    radius: float = 0.5,
    offset: Vector = np.array([0.0, 0.5]),
) -> Trajectory:
    t_0, t_f = t_range

    def trajectory(t: float) -> Vector:
        normalized_t = (t - t_0) / (t_f - t_0) * 2.0 * np.pi
        x = radius * np.sin(normalized_t)
        y = radius * np.sin(normalized_t) * np.cos(normalized_t)

        return np.array([x, y]) + offset + radius

    return trajectory


def lemniscate(
    t_range: tuple[float, float],
    *,
    radius: float = 0.5,
    offset: Vector = np.array([0.0, 0.5]),
) -> Trajectory:
    t_0, t_f = t_range

    def trajectory(t: float) -> Vector:
        normalized_t = (t - t_0) / (t_f - t_0) * 2.0 * np.pi
        x = radius * np.sin(normalized_t)
        y = radius * np.sin(2 * normalized_t)

        return np.array([x, y]) + offset + radius

    return trajectory


def clover(
    t_range: tuple[float, float],
    *,
    radius: float = 0.5,
    offset: Vector = np.array([0.0, 0.5]),
) -> Trajectory:
    t_0, t_f = t_range

    def trajectory(t: float) -> Vector:
        normalized_t = (t - t_0) / (t_f - t_0) * 2.0 * np.pi
        x = radius * np.cos(normalized_t) * np.cos(2 * normalized_t)
        y = radius * np.sin(normalized_t) * np.cos(2 * normalized_t)

        return np.array([x, y]) + offset + radius

    return trajectory


def line(
    t_range: tuple[float, float],
    *,
    start: tuple[float, float],
    end: tuple[float, float],
) -> Trajectory:
    t_0, t_f = t_range
    delta_t = t_f - t_0
    x_0, y_0 = start
    x_f, y_f = end

    def trajectory(t: float) -> Vector:
        normalized_t = (t - t_0) / delta_t
        x = x_0 + normalized_t * (x_f - x_0)
        y = y_0 + normalized_t * (y_f - y_0)

        return np.array([x, y])

    return trajectory



from typing import Protocol, Sequence, Generic, Callable

from helpers.tank.policy import Policy, StateT, ActionT, State, RewardFunction
from helpers.tank.learn import TrainingResult

from plotly.graph_objects import Figure, Scatter, Bar
from plotly.subplots import make_subplots


import numpy as np


class PolicyVisualizationAddOn(Protocol, Generic[ActionT]):
    def __call__(
        self, figure: Figure, actions: Sequence[tuple[ActionT, float]]
    ) -> None:
        """Adds additional information to the policy visualization."""
        ...


def no_add_on(figure: Figure, actions: Sequence[tuple[ActionT, float]]) -> None:
    """Does nothing."""
    pass


def visualize_discretization(
    discretization: Sequence[State],
    *,
    min_value: float | None = None,
    max_value: float,
    state_name: str,
    unit: str | None = None,
    title: str,
) -> None:
    """
    Visualizes the discretization of a system state.

    Args:
        discretization: A sequence of states that have been discretized.
        min_value: The minimum value of the state (optional).
        max_value: The maximum value of the state.
        state_name: The name of the state.
        unit: The unit of the state (if it is a physical quantity).
        title: The title of the plot.
    """
    discrete_values = [state.value for state in discretization]

    figure = Figure()

    figure.add_vline(
        x=0,
        line_dash="dash",
        line_color="gray",
        annotation_text=f"No {state_name}",
        annotation_position="top",
    )

    if min_value is not None:
        figure.add_vline(
            x=min_value,
            line_dash="dash",
            line_color="red",
            annotation_text=f"Min {state_name}",
            annotation_position="bottom",
            annotation=dict(yshift=-20),
        )

    figure.add_vline(
        x=max_value,
        line_dash="dash",
        line_color="red",
        annotation_text=f"Max {state_name}",
        annotation_position="bottom",
        annotation=dict(yshift=-20),
    )

    figure.add_trace(
        Scatter(
            x=discrete_values,
            y=[0] * len(discrete_values),
            mode="markers",
            name="Discrete States",
            marker=dict(size=12, symbol="diamond-tall"),
        )
    )

    # Update layout
    figure.update_layout(
        title=title,
        xaxis_title=f"{state_name} ({unit})" if unit else state_name,
        yaxis_showticklabels=False,
        yaxis_zeroline=False,
        showlegend=True,
        yaxis_range=[-0.5, 0.5],
    )

    figure.show()


def visualize_reward_function(
    reward_function: RewardFunction[StateT],
    states: Sequence[StateT],
    state_from: Callable[[float], StateT],
    *,
    state_name: str,
    unit: str | None = None,
    samples: int = 2000,
    title: str = "Tank System Reward Function",
) -> None:
    """
    Visualizes the reward function and discretized states.

    Args:
        reward_function: The reward function to visualize.
        states: Sequence of states following the State protocol.
        state_from: A function that creates a state from a numerical value.
        state_name: The name of the state.
        unit: The unit of the state (if it is a physical quantity).
        samples: Number of points for smooth reward function curve.
        title: Title of the plot.
    """
    discrete_values = [state.value for state in states]
    continuous_values = np.linspace(min(discrete_values), max(discrete_values), samples)

    rewards_discrete = [reward_function(state) for state in states]
    rewards_continuous = [
        reward_function(state_from(value)) for value in continuous_values
    ]

    figure = Figure()

    figure.add_trace(
        Scatter(
            x=discrete_values,
            y=rewards_discrete,
            mode="markers",
            name="Discrete States",
            marker=dict(size=12, symbol="diamond-tall"),
        )
    )

    figure.add_trace(
        Scatter(
            x=continuous_values,
            y=rewards_continuous,
            mode="lines",
            name="Reward Function",
            line=dict(width=2),
            opacity=0.6,
        )
    )

    figure.update_layout(
        title=title,
        xaxis_title=f"{state_name} ({unit})" if unit else state_name,
        yaxis_title="Reward",
        showlegend=True,
    )

    figure.show()


def visualize_policy(
    policy: Policy[ActionT],
    actions: Sequence[tuple[ActionT, float]],
    *,
    title: str,
    samples: int = 1000,
    add_on: PolicyVisualizationAddOn[ActionT] = no_add_on,
) -> None:
    """
    Visualizes epsilon-greedy policy decisions.

    Args:
        actions: Sequence of (action, value) pairs.
        epsilon: Exploration parameter.
        samples: Number of policy samples to simulate.
        title: Plot title.
    """
    figure = make_subplots(specs=[[{"secondary_y": True}]])

    actions = sorted(actions, key=lambda x: x[1], reverse=True)
    action_names = [str(action) for action, _ in actions]
    values = [value for _, value in actions]

    # Simulate policy decisions
    selections = [policy(actions) for _ in range(samples)]
    selection_counts = [selections.count(action[0]) / samples for action in actions]

    # Add value bars
    figure.add_trace(
        Bar(
            x=action_names,
            y=values,
            name="Action Values",
            opacity=0.6,
        ),
        secondary_y=False,
    )

    figure.add_trace(
        Scatter(
            x=action_names,
            y=selection_counts,
            name="Selection Probability",
            mode="lines+markers",
            marker=dict(size=10),
            line=dict(width=2),
        ),
        secondary_y=True,
    )

    add_on(figure, actions)

    figure.update_layout(
        title=title, xaxis_title="Actions", showlegend=True, height=500
    )

    figure.update_yaxes(title_text="Action Value", secondary_y=False)
    figure.update_yaxes(
        title_text="Selection Probability", secondary_y=True, range=[0, 1]
    )

    figure.show()


def theoretical_epsilon_greedy_probabilities(
    epsilon: float,
) -> PolicyVisualizationAddOn:
    """Returns a policy visualization add-on that includes theoretical probabilities.

    Args:
        epsilon: Exploration parameter.

    Returns:
        A policy visualization add-on that includes theoretical probabilities.
    """

    def add_on(figure: Figure, actions: Sequence[tuple[ActionT, float]]) -> None:
        actions = sorted(actions, key=lambda x: x[1], reverse=True)
        action_names = [str(action) for action, _ in actions]

        random_probability = epsilon / len(actions)
        best_action_probability = 1 - epsilon + random_probability
        theoretical_probs = [
            best_action_probability if i == 0 else random_probability
            for i in range(len(actions))
        ]

        figure.add_trace(
            Scatter(
                x=action_names,
                y=theoretical_probs,
                name="Theoretical Probability",
                mode="lines+markers",
                line=dict(width=2, dash="dash"),
                marker=dict(size=10, symbol="x"),
            ),
            secondary_y=True,
        )

    return add_on


def visualize_training_results(results: Sequence[TrainingResult]) -> None:
    """Visualizes training results in a 4x1 subplot layout.

    Args:
        results: Sequence of training results.
    """

    figure = make_subplots(
        rows=4,
        cols=1,
        subplot_titles=(
            "Tank Heights",
            "Learning Rate",
            "Height Difference",
            "Accumulated Reward",
        ),
        vertical_spacing=0.1,
    )

    # Heights plot (starting height and setpoint)
    figure.add_trace(
        Scatter(
            y=[r.starting_height for r in results],
            name="Starting Height",
            line=dict(color="blue"),
        ),
        row=1,
        col=1,
    )

    figure.add_trace(
        Scatter(
            y=[r.setpoint for r in results],
            name="Setpoint",
            line=dict(color="red", dash="dash"),
        ),
        row=1,
        col=1,
    )

    # Learning rate
    figure.add_trace(
        Scatter(
            y=[r.learning_rate for r in results],
            name="Learning Rate",
            line=dict(color="green"),
        ),
        row=2,
        col=1,
    )

    # Height difference
    figure.add_trace(
        Scatter(
            y=[r.height_difference for r in results],
            name="Height Difference",
            line=dict(color="purple"),
        ),
        row=3,
        col=1,
    )

    # Add zero line for height difference
    figure.add_hline(
        y=0,
        line_dash="dash",
        line_color="gray",
        row=3,  # type: ignore
        col=1,  # type: ignore
    )

    # Accumulated reward
    figure.add_trace(
        Scatter(
            y=[r.accumulated_reward for r in results],
            name="Accumulated Reward",
            line=dict(color="orange"),
        ),
        row=4,
        col=1,
    )

    # Update layout
    figure.update_layout(
        height=1000, showlegend=True, title_text="Training Results Over Time"
    )

    figure.update_yaxes(title_text="Height (m)", row=1, col=1)
    figure.update_yaxes(title_text="Learning Rate", row=2, col=1)
    figure.update_yaxes(title_text="Height Diff (m)", row=3, col=1)
    figure.update_yaxes(title_text="Reward", row=4, col=1)
    figure.update_xaxes(title_text="Episode", row=4, col=1)

    figure.show()



from typing import Protocol
from dataclasses import dataclass

from helpers.maths import Matrix, Vector
from helpers.tank.control import Control
from helpers.tank.simulate import (
    TankDynamics,
    SimulationResults,
    IntermediateValueRecorder,
    TankT,
)


import numpy as np


class Tank(Protocol):
    def inlet_flow_rate_for(self, control: float) -> float:
        """Returns the inlet flow rate for the given control value."""
        ...

    def outlet_flow_rate_for(self, height: float) -> float:
        """Returns the outlet flow rate for the given liquid height."""
        ...

    @property
    def area(self) -> float:
        """Returns the cross-sectional area of the tank."""
        ...

    @property
    def liquid_height(self) -> float:
        """Returns the current height of the liquid in the tank."""
        ...

    @property
    def height(self) -> float:
        """Returns the total height of the tank."""
        ...


@dataclass(frozen=True)
class SimpleTankDynamics:
    tank: Tank
    control: Control
    recorder: IntermediateValueRecorder
    limiting_factor: float = 10.0

    def __call__(self, t: float, state: Vector) -> Vector:
        """Returns the derivative of the tank height.

        Note:
            The state vector contains just the liquid height in the tank.
            The derivative is calculated directly from inflow - outflow.
        """
        h = state[0]
        inlet_flow_rate = self.tank.inlet_flow_rate_for(self.control(t=t, h=h))
        outlet_flow_rate = self.tank.outlet_flow_rate_for(h)

        # The inlet flow rate is limited by the available space in the tank. We
        # use a smooth function to limit the flow rate to the maximum possible value.
        space_left = self.tank.height - h
        inlet_flow_rate = inlet_flow_rate * np.tanh(self.limiting_factor * space_left)

        self.recorder.record(
            t=t, inlet_flow_rate=inlet_flow_rate, outlet_flow_rate=outlet_flow_rate
        )

        dh_dt = (inlet_flow_rate - outlet_flow_rate) / self.tank.area

        return np.array([dh_dt])

    def jacobian(self) -> None:
        return None


class TankDynamicsMixin:
    def dynamics_for(
        self: Tank, control: Control, recorder: IntermediateValueRecorder
    ) -> TankDynamics:
        return SimpleTankDynamics(self, control, recorder)

    def initial_conditions(self: Tank) -> Vector:
        return np.array([self.liquid_height])

    def results_for(
        self: TankT,
        t_range: tuple[float, float],
        recorded_points: int,
        t: Vector,
        inflow_rate: Vector,
        outflow_rate: Vector,
        solution: Matrix,
    ) -> SimulationResults[TankT]:
        return SimulationResults(
            tank=self,
            time=t,
            inlet_flow_rate=inflow_rate,
            outlet_flow_rate=outflow_rate,
            liquid_height=solution[0],
        )



from typing import Protocol, Generic, TypeVar, Sequence, Self, overload
from dataclasses import dataclass, field


from helpers.maths import Matrix
from numpy.random import Generator

import numpy as np


StateT = TypeVar("StateT", bound="State", infer_variance=True)
ActionT = TypeVar("ActionT", bound="Action", infer_variance=True)


class Sortable(Protocol):
    def __lt__(self, other: Self) -> bool:
        """Returns whether this instance is less than the other instance."""
        ...


class Hashable(Protocol):
    def __hash__(self) -> int:
        """Returns the hash value of this instance."""
        ...

    def __eq__(self, other: object) -> bool:
        """Returns whether this instance is equal to the other instance."""
        ...


class State(Sortable, Hashable, Protocol):
    @property
    def value(self) -> float:
        """Returns the value of the state."""
        ...


class Action(Sortable, Hashable, Protocol):
    @property
    def value(self) -> float:
        """Returns the value of the action."""
        ...


class QTable(Protocol, Generic[StateT, ActionT]):
    @overload
    def get(self, state: StateT) -> Sequence[tuple[ActionT, float]]:
        """Returns all possible actions and their Q-values for the given state.

        Args:
            state: The state for which to get the Q-values.

        Returns:
            A sequence of tuples, where each tuple contains an action and its corresponding Q-value.

        Example:
            You can get the actions and Q-values for a state like this:

            ```python
            state = ... # Some state
            actions = q_table.get(state)

            print(actions)
            # This will print something like this:
            # [(action_1, q_value_1), (action_2, q_value_2), ...]
            ```
        """
        ...

    @overload
    def get(self, state: StateT, action: ActionT) -> float:
        """Returns the Q-value for the given state-action pair.

        Args:
            state: The state for which to get the Q-value.
            action: The action for which to get the Q-value.

        Returns:
            The Q-value for the given state-action pair.

        Example:
            You can get the Q-value for a state-action pair like this:

            ```python
            state = ... # Some state
            action = ... # Some action
            q_value = q_table.get(state, action)

            print(q_value)
            # This will print the Q-value for the state-action pair, like this:
            # 42.0
            ```
        """
        ...

    def update(self, state: StateT, action: ActionT, *, q_value: float) -> None:
        """Updates the Q-value for the given state-action pair.

        Args:
            state: The state for which to update the Q-value.
            action: The action for which to update the Q-value.
            q_value: The new Q-value for the state-action pair.

        Example:
            You can update the Q-value for a state-action pair like this:

            ```python
            state = ... # Some state
            action = ... # Some action
            new_q_value = ... # Some new Q-value
            q_table.update(state, action, q_value=new_q_value)
            ```
        """
        ...


class QValueInitializer(Protocol):
    def initialize(self, *, state_count: int, action_count: int) -> Matrix:
        """Creates a matrix of Q-values with the given dimensions."""
        ...


class ZeroInitializer:
    def initialize(self, *, state_count: int, action_count: int) -> Matrix:
        return np.zeros((state_count, action_count))


@dataclass(frozen=True)
class ConstantInitializer:
    value: float

    def initialize(self, *, state_count: int, action_count: int) -> Matrix:
        return np.full((state_count, action_count), self.value)


@dataclass(frozen=True)
class RandomInitializer:
    min_value: float = 0.0
    max_value: float = 1.0
    random: Generator = field(default_factory=np.random.default_rng)

    def initialize(self, *, state_count: int, action_count: int) -> Matrix:
        return self.random.uniform(
            self.min_value, self.max_value, (state_count, action_count)
        )


@dataclass(frozen=True)
class SimpleQTable(Generic[StateT, ActionT]):
    table: dict[StateT, dict[ActionT, float]]
    states: Sequence[StateT]
    actions: Sequence[ActionT]

    delta_state: float
    delta_action: float

    @staticmethod
    def create_for(
        *,
        states: Sequence[StateT],
        actions: Sequence[ActionT],
        initializer: QValueInitializer,
        delta_state: float,
        delta_action: float,
    ) -> "SimpleQTable":
        """Creates a simple Q-table with all Q-values initialized"""
        states = sorted(states)
        actions = sorted(actions)
        q_values = initializer.initialize(
            state_count=len(states), action_count=len(actions)
        )

        return SimpleQTable._from_q_values(
            q_values,
            states=states,
            actions=actions,
            delta_state=delta_state,
            delta_action=delta_action,
        )

    @staticmethod
    def from_file(
        file: str,
        *,
        states: Sequence[StateT],
        actions: Sequence[ActionT],
        delta_state: float,
        delta_action: float,
        verify: bool = True,
    ) -> "SimpleQTable":
        array = np.loadtxt(file)
        q_values = np.zeros((len(states), len(actions)))

        for i, row in enumerate(array):
            for j, (state, action, q_value) in enumerate(row):
                if verify:
                    assert state == states[i].value, (
                        f"Incorrect state at row {i}. Expected the state value "
                        f"to be {states[i].value}, but got {state}."
                    )
                    assert action == actions[j].value, (
                        f"Incorrect action at column {j}. Expected the action value "
                        f"to be {actions[j].value}, but got {action}."
                    )

                q_values[i, j] = q_value

        return SimpleQTable._from_q_values(
            q_values,
            states=states,
            actions=actions,
            delta_state=delta_state,
            delta_action=delta_action,
        )

    @overload
    def get(self, state: StateT) -> Sequence[tuple[ActionT, float]]: ...

    @overload
    def get(self, state: StateT, action: ActionT) -> float: ...

    def get(
        self, state: StateT, action: ActionT | None = None
    ) -> Sequence[tuple[ActionT, float]] | float:
        if action is None:
            return tuple(
                (action, value)
                for action, value in self.table[self._state_for(state)].items()
            )
        else:
            return self.table[self._state_for(state)][self._action_for(action)]

    def update(self, state: StateT, action: ActionT, *, q_value: float) -> None:
        self.table[self._state_for(state)][self._action_for(action)] = q_value

    def save_to(self, path: str) -> None:
        array = np.array(
            [
                [state.value, action.value, q_value]
                for state, actions in self.table.items()
                for action, q_value in actions.items()
            ]
        )

        np.savetxt(path, array)

    @staticmethod
    def _from_q_values(
        q_values: Matrix,
        *,
        states: Sequence[StateT],
        actions: Sequence[ActionT],
        delta_state: float,
        delta_action: float,
    ) -> "SimpleQTable":
        return SimpleQTable(
            table={
                state: {action: q_values[i, j] for j, action in enumerate(actions)}
                for i, state in enumerate(states)
            },
            states=states,
            actions=actions,
            delta_state=delta_state,
            delta_action=delta_action,
        )

    def _state_for(self, state: StateT) -> StateT:
        min_state = self.states[0].value
        index = int((state.value - min_state) / self.delta_state)
        index = np.clip(index, 0, len(self.states) - 1)

        return self.states[index]

    def _action_for(self, action: ActionT) -> ActionT:
        min_action = self.actions[0].value
        index = int((action.value - min_action) / self.delta_action)
        index = np.clip(index, 0, len(self.actions) - 1)

        return self.actions[index]



from dataclasses import dataclass, KW_ONLY, replace


@dataclass(frozen=True)
class Joint:
    """A joint in a robot arm.

    A joint connects two links and allows them to rotate relative to each other. The joint is
    assumed to be massless, but it can have a certain friction that resists the movement of the
    links.

    Attributes:
        theta: The angle of the joint.
        friction: The friction of the joint.

    Example:
        When creating a [`PlanarRobotArm`](exercise_3.md#helpers.robot.PlanarRobotArm), you can define
        the joints of the robot arm as follows:

        ```python
        robot = PlanarRobotArm(
            joint_1=Joint(theta=0.0, friction=0.1),
            joint_2=Joint(theta=pi / 2, friction=0.2),
            ... # Other parameters
        )

        # This will create a planar robot arm with two joints.
        ```
    """

    _: KW_ONLY
    theta: float
    friction: float = 0.01

    def __post_init__(self) -> None:
        assert 0 <= self.friction, (
            f"The friction of the joint must be greater than or equal to zero. "
            f"Got {self.friction}."
        )

    def with_angle(self, theta: float) -> "Joint":
        """Returns a new joint with the given angle.

        Args:
            theta: The angle of the joint.

        Returns:
            A new joint with the given angle.
        """
        return replace(self, theta=theta)



from typing import Protocol, Generic, TypeVar


StateT = TypeVar("StateT", infer_variance=True)
ActionT = TypeVar("ActionT", infer_variance=True)


class LearningAgent(Protocol, Generic[StateT, ActionT]):
    def __call__(self, state: StateT) -> ActionT:
        """Returns the action to take for the given state."""
        ...

    def update(
        self, state: StateT, action: ActionT, next_state: StateT, reward: float
    ) -> None:
        """Updates the learning agent based on the observed transition."""
        ...

    @property
    def learning_rate(self) -> float:
        """Returns the learning rate of the agent."""
        ...

    @learning_rate.setter
    def learning_rate(self, value: float) -> None:
        """Sets the learning rate of the agent."""
        ...

    @property
    def discount_factor(self) -> float:
        """Returns the discount factor of the agent."""
        ...

    @discount_factor.setter
    def discount_factor(self, value: float) -> None:
        """Sets the discount factor of the agent."""
        ...



from . import (
    task_space as task_space,
    joint_space as joint_space,
    trajectories as trajectories,
)
from .control import (
    zero as zero,
    chirp as chirp,
    constant as constant,
    ramp as ramp,
    step as step,
    SetpointFollower as SetpointFollower,
    Trajectory as Trajectory,
    ControlSignalProvider as ControlSignalProvider,
)



from typing import Final, override, Sequence, Any, TypeAlias, Protocol
from dataclasses import dataclass, KW_ONLY, replace
from helpers.maths import Point
from helpers.robot.animation import (
    AnimatableRobotArm,
    JointAngles,
    RobotAnimator,
    DoNothing,
)
from helpers.robot.arm import RobotArmMixin
from helpers.robot.link import Link

import numpy as np
from plotly.graph_objects import Figure, Scatter

LINK_WIDTH: Final[float] = 0.15

Styling: TypeAlias = dict[str, Any]


@dataclass(frozen=True)
class StylingOverrides:
    _: KW_ONLY
    opacity: float
    trace_color: str
    name: str | None

    @staticmethod
    def none() -> "StylingOverrides":
        return StylingOverrides(opacity=1.0, trace_color="LightBlue", name=None)

    def __post_init__(self):
        assert 0 <= self.opacity <= 1, (
            f"Opacity must be between 0 and 1. Got {self.opacity}."
        )

    def fill_name(self, name: str) -> "StylingOverrides":
        return replace(self, name=name) if self.name is None else self


class AnimatablePlanarRobotArm(AnimatableRobotArm, Protocol):
    @override
    def draw(
        self,
        *,
        show: bool = True,
        trace: Sequence[Point] = (),
        overrides: StylingOverrides = StylingOverrides.none(),
    ) -> Figure:
        """Draws the robot arm in two dimensions."""
        ...

    @property
    def link_1(self) -> Link:
        """The first link of the robot arm."""
        ...

    @property
    def link_2(self) -> Link:
        """The second link of the robot arm."""
        ...

    @override
    def _draw_trace(
        self,
        trace: Sequence[Point],
        figure: Figure,
        *,
        overrides: StylingOverrides = StylingOverrides.none(),
    ) -> Figure:
        """Draws the trace of the end effector of the robot arm on the given figure."""
        ...

    @override
    def _draw_robot(
        self,
        joint_positions: Sequence[Point],
        figure: Figure,
        *,
        overrides: StylingOverrides = StylingOverrides.none(),
    ) -> Figure:
        """Draws the robot arm on the given figure."""
        ...

    def _draw_joints(
        self,
        joint_positions: Sequence[Point],
        figure: Figure,
        *,
        overrides: StylingOverrides,
    ) -> Figure:
        """Draws the joints of the robot arm on the given figure."""
        ...

    def _draw_links(
        self,
        joint_positions: Sequence[Point],
        figure: Figure,
        *,
        overrides: StylingOverrides,
    ) -> Figure:
        """Draws the links of the robot arm on the given figure."""
        ...

    def _draw_centers_of_mass(
        self,
        centers_of_mass: Sequence[Point],
        figure: Figure,
        *,
        overrides: StylingOverrides,
    ) -> Figure:
        """Draws the centers of mass of the links of the robot arm on the given figure."""
        ...

    def _trace_style(self, *, overrides: StylingOverrides) -> Styling:
        """The style of the trace of the end effector of the robot arm."""
        ...

    def _joint_style(self, *, overrides: StylingOverrides) -> Styling:
        """The style of the joints of the robot arm."""
        ...

    def _link_style(self, *, overrides: StylingOverrides) -> Styling:
        """The style of the links of the robot arm."""
        ...

    def _center_of_mass_style(self, *, overrides: StylingOverrides) -> Styling:
        """The style of the centers of mass of the robot arm."""
        ...

    @property
    def _links(self) -> list[Link]:
        """The links of the robot arm."""
        ...


class PlanarRobotAnimationMixin(RobotArmMixin):
    """This mixin overrides some details of [`RobotArmMixin`](exercise_2.md#helpers.robot.RobotArmMixin) on
    how the robot arm is visualized during animations. It also provides an `animate` method directly
    on the robot arm object, without the need to use the [`RobotAnimator`](exercise_2.md#helpers.robot.RobotAnimator)
    class.

    Example:
        ```python
        robot = ... # Some robot arm object that uses the `PlanarRobotAnimationMixin`
        joint_angles = [(0, 0), (np.pi / 2, np.pi / 2), (np.pi, np.pi)]

        robot.animate(joint_angles)
        # This will animate the robot arm moving through the given joint angles.
        ```
    """

    def animate(
        self: AnimatablePlanarRobotArm,
        joint_angles: list[JointAngles],
        *,
        predicted_joint_angles: list[JointAngles] | None = None,
        align_starting_position: bool = False,
        subtitle: str = "",
    ) -> None:
        """Animates the robot arm moving through the given joint angles. See the
        [`animate`](exercise_2.md#helpers.robot.RobotAnimator.animate) function for more details.
        """
        RobotAnimator.using(PlanarRobotAnimationFramesProvider()).animate(
            self,
            joint_angles,
            align_starting_position=align_starting_position,
            subtitle=subtitle,
            on_draw=DrawPrediction.of(
                self,
                predicted_joint_angles,
                opacity=0.35,
                trace_color="rgba(255, 183, 77, 0.5)",
            )
            if predicted_joint_angles is not None
            else DoNothing(),
        )

    @override
    def draw(
        self: AnimatablePlanarRobotArm,
        *,
        show: bool = True,
        trace: Sequence[Point] = (),
        overrides: StylingOverrides = StylingOverrides.none(),
    ) -> Figure:
        overrides = overrides.fill_name(self.name)
        joint_positions = self.joint_positions()

        figure = Figure()

        if trace:
            figure = self._draw_trace(trace, figure, overrides=overrides)

        figure = self._draw_robot(joint_positions, figure, overrides=overrides)

        self._set_layout_for(figure)

        if show:
            figure.show()

        return figure

    @override
    def _draw_trace(
        self: AnimatablePlanarRobotArm,
        trace: Sequence[Point],
        figure: Figure,
        *,
        overrides: StylingOverrides = StylingOverrides.none(),
    ) -> Figure:
        figure.add_trace(
            Scatter(
                x=[point.x for point in trace],
                y=[point.y for point in trace],
                **self._trace_style(overrides=overrides),
            ),
        )

        return figure

    @override
    def _draw_robot(
        self: AnimatablePlanarRobotArm,
        joint_positions: Sequence[Point],
        figure: Figure,
        *,
        overrides: StylingOverrides = StylingOverrides.none(),
    ) -> Figure:
        figure = self._draw_joints(joint_positions, figure, overrides=overrides)
        figure = self._draw_links(joint_positions, figure, overrides=overrides)

        return figure

    def _draw_joints(
        self: AnimatablePlanarRobotArm,
        joint_positions: Sequence[Point],
        figure: Figure,
        *,
        overrides: StylingOverrides,
    ) -> Figure:
        figure.add_trace(
            Scatter(
                x=[point.x for point in joint_positions],
                y=[point.y for point in joint_positions],
                **self._joint_style(overrides=overrides),
            ),
        )

        return figure

    def _draw_links(
        self: AnimatablePlanarRobotArm,
        joint_positions: Sequence[Point],
        figure: Figure,
        *,
        overrides: StylingOverrides,
    ) -> Figure:
        link_width = LINK_WIDTH
        centers_of_mass: list[Point] = []

        for i, (p_1, p_2, link) in enumerate(
            zip(joint_positions, joint_positions[1:], self._links)
        ):
            # Calculate positions of the rectangle corners
            angle = np.arctan2(p_2.y - p_1.y, p_2.x - p_1.x)
            dx = link_width / 2 * np.sin(angle)
            dy = -link_width / 2 * np.cos(angle)

            points = [
                Point(p_1.x - dx, p_1.y - dy),
                Point(p_1.x + dx, p_1.y + dy),
                Point(p_2.x + dx, p_2.y + dy),
                Point(p_2.x - dx, p_2.y - dy),
            ]

            # Calculate the center of mass of the link
            centers_of_mass.append(
                Point(
                    x=p_1.x + link.distance_to_com * np.cos(angle),
                    y=p_1.y + link.distance_to_com * np.sin(angle),
                )
            )

            figure.add_trace(
                Scatter(
                    # The extra point at the end closes the rectangle
                    x=[p.x for p in points] + [points[0].x],
                    y=[p.y for p in points] + [points[0].y],
                    showlegend=i == 0,
                    **self._link_style(overrides=overrides),
                ),
            )

        figure = self._draw_centers_of_mass(
            centers_of_mass, figure, overrides=overrides
        )

        return figure

    def _draw_centers_of_mass(
        self: AnimatablePlanarRobotArm,
        centers_of_mass: Sequence[Point],
        figure: Figure,
        *,
        overrides: StylingOverrides,
    ) -> Figure:
        figure.add_trace(
            Scatter(
                x=[point.x for point in centers_of_mass],
                y=[point.y for point in centers_of_mass],
                **self._center_of_mass_style(overrides=overrides),
            ),
        )

        return figure

    def _trace_style(
        self: AnimatablePlanarRobotArm, *, overrides: StylingOverrides
    ) -> Styling:
        name, trace_color = overrides.name, overrides.trace_color

        return dict(
            name=f"{name}'s trace",
            mode="lines",
            line=dict(width=2, color=trace_color, dash="dash"),
        )

    def _joint_style(
        self: AnimatablePlanarRobotArm, *, overrides: StylingOverrides
    ) -> Styling:
        opacity, name = overrides.opacity, overrides.name

        return dict(
            name="Joints",
            legendgroup=name,
            legendgrouptitle=dict(text=name),
            mode="markers",
            marker=dict(
                size=10,
                symbol="circle",
                color=f"rgba(0, 0, 255, {opacity})",
                line=dict(width=1, color=f"rgba(0, 0, 0, {opacity})"),
            ),
        )

    def _link_style(
        self: AnimatablePlanarRobotArm, *, overrides: StylingOverrides
    ) -> Styling:
        opacity, name = overrides.opacity, overrides.name

        return dict(
            name="Links",
            legendgroup=name,
            legendgrouptitle=dict(text=name),
            mode="lines",
            line=dict(color=f"rgba(0, 0, 255, {0.75 * opacity})", width=1),
            fill="toself",
            fillcolor=f"rgba(0, 0, 255, {0.15 * opacity})",
        )

    def _center_of_mass_style(
        self: AnimatablePlanarRobotArm, *, overrides: StylingOverrides
    ) -> Styling:
        opacity, name = overrides.opacity, overrides.name

        return dict(
            name="Centers of Mass",
            legendgroup=name,
            legendgrouptitle=dict(text=name),
            mode="markers",
            marker=dict(
                size=10,
                symbol="circle-cross",
                color=f"rgba(0, 0, 255, {0.15 * opacity})",
                line=dict(width=1, color=f"rgba(0, 0, 0, {0.75 * opacity})"),
            ),
        )

    @property
    def _links(self: AnimatablePlanarRobotArm) -> list[Link]:
        return [self.link_1, self.link_2]


@dataclass(frozen=True)
class PlanarRobotAnimationFramesProvider:
    _: KW_ONLY
    overrides: StylingOverrides = StylingOverrides.none()

    def __call__(
        self, robot: AnimatablePlanarRobotArm, joint_angles: list[JointAngles]
    ) -> list[Figure]:
        end_effector_positions: list[Point] = []
        frames: list[Figure] = []

        for theta_1, theta_2 in joint_angles:
            robot.rotate_to(theta_1, theta_2)

            end_effector_positions.append(robot.end_effector_position())
            frames.append(
                robot.draw(
                    show=False,
                    trace=end_effector_positions,
                    overrides=self.overrides,
                )
            )

        return frames


@dataclass
class DrawPrediction:
    prediction_frames: list[Figure]
    current_frame: int = 0

    @staticmethod
    def of(
        robot: AnimatablePlanarRobotArm,
        predicted_joint_angles: list[JointAngles],
        *,
        opacity: float,
        trace_color: str,
    ) -> "DrawPrediction":
        return DrawPrediction(
            PlanarRobotAnimationFramesProvider(
                overrides=StylingOverrides(
                    opacity=opacity,
                    trace_color=trace_color,
                    name=f"(Predicted) {robot.name}",
                )
            )(robot, predicted_joint_angles)
        )

    def __call__(self, figure: Figure) -> Figure:
        assert self.current_frame < len(self.prediction_frames), (
            "No more prediction frames to draw."
        )

        figure.add_traces(self.prediction_frames[self.current_frame].data)
        self.current_frame += 1

        return figure



from . import linear as linear, solver as solver
from .latex import (
    pretty as pretty,
    pretty_latex as pretty_latex,
)
from .matrix import MatrixPowers as MatrixPowers
from .types import (
    Matrix as Matrix,
    Vector as Vector,
    IntMatrix as IntMatrix,
    IntVector as IntVector,
    BoolMatrix as BoolMatrix,
    Point as Point,
)



from typing import Sequence, Generic, Protocol, TypeVar, Self
from dataclasses import dataclass, KW_ONLY

from helpers.maths import Vector, Matrix
from helpers.robot.animation import JointAngles
from helpers.robot.simulate import (
    SimulatableRobotArm,
    ControlSignal,
    StatefulControlSignal,
    SimulationResults,
    RobotDynamics,
)


SimulatableRobotT = TypeVar(
    "SimulatableRobotT", bound="SimulatableRobotArm", infer_variance=True
)
PredictableRobotT = TypeVar(
    "PredictableRobotT", bound="PredictableRobotArm", covariant=True
)


class PredictableRobotArm(SimulatableRobotArm, Protocol):
    def using_dynamics_from(self, provider: "RobotDynamicsProvider") -> Self:
        """Returns a new robot arm that uses the dynamics model from the given provider."""
        ...


class AnimatedRobotArm(SimulatableRobotArm, Protocol):
    def animate(
        self,
        joint_angles: list[JointAngles],
        *,
        predicted_joint_angles: list[JointAngles] | None = None,
    ) -> None:
        """Animates the robot arm moving through the given joint angles. See the
        [`animate`](exercise_2.md#helpers.robot.RobotAnimator.animate) function for more details.

        Note:
            If `predicted_joint_angles` is provided, the animation should show both the actual and predicted
            joint angles overlaid on top of each other.
        """
        ...


class RobotDynamicsProvider(Protocol, Generic[SimulatableRobotT]):
    def create_for(
        self, robot: SimulatableRobotT, *, torque: ControlSignal
    ) -> RobotDynamics:
        """Creates an object that can simulate the dynamics of the robot for the given torque."""
        ...

    def results_for(
        self,
        robot: SimulatableRobotT,
        *,
        t_range: tuple[float, float],
        recorded_points: int,
        t: Vector,
        tau: list[Vector],
        solution: Matrix,
    ) -> SimulationResults[SimulatableRobotT]:
        """Extracts the simulation results from the specified state matrix."""
        ...


@dataclass(frozen=True)
class PredictionResults(Generic[SimulatableRobotT]):
    _: KW_ONLY
    actual: SimulationResults[SimulatableRobotT]
    predicted: SimulationResults[SimulatableRobotT]

    def animate(
        self: "PredictionResults[AnimatedRobotArm]", *, take: int = 1
    ) -> "PredictionResults[AnimatedRobotArm]":
        actual_angles = angles_from(self.actual)[::take]
        predicted_angles = angles_from(self.predicted)[::take]

        self.actual.robot.animate(
            actual_angles, predicted_joint_angles=predicted_angles
        )

        return self


class RobotArmPredictionMixin:
    def predict(
        self: PredictableRobotT,
        tau: ControlSignal | StatefulControlSignal,
        *,
        model: RobotDynamicsProvider,
        t_range: tuple[float, float],
        recorded_points: int | None = None,
        initial_velocities: Sequence[float] | None = None,
        initial_conditions: Sequence[float] | Vector | None = None,
    ) -> PredictionResults[PredictableRobotT]:
        predictor = self.using_dynamics_from(model)

        predicted_results = predictor.simulate(
            tau,
            t_range=t_range,
            recorded_points=recorded_points,
            initial_velocities=initial_velocities,
            initial_conditions=initial_conditions,
        )

        actual_results = self.simulate(
            tau,
            t_range=t_range,
            recorded_points=recorded_points,
            initial_velocities=initial_velocities,
            initial_conditions=initial_conditions,
        )

        return PredictionResults(actual=actual_results, predicted=predicted_results)


def angles_from(results: SimulationResults) -> list[JointAngles]:
    return JointAngles.combining(results.q[0], results.q[1])
